#
# destables
# Copyright (C) 2022  Arthur Wisz
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
#

import PySimpleGUI as sg
from commonwin import *
import clipboard
import tabhelper
from tabulate import tabulate
from c11tools import *

TYPE_NAME = 'Type name:'
DESCRIPTION = 'Description:'
KIND = 'Kind:'
HEADER = 'Declared in:'
TYPEDEF = 'Type:'
ELEMENTS = 'Elements:'
CONSTANTS = 'Constants:'

def show_type_table(indent_size, add_table_directive):

    kinds = [ 'Typedef', 'Structure', 'Enumeration' ]

    inputs = [
        sg.InputText(key=TYPE_NAME, size=(64, 1)),
        sg.Multiline(key=DESCRIPTION, enter_submits=False, autoscroll=True, size=(64, 2)),
        sg.Combo(kinds, kinds[0], key=KIND, enable_events=True, readonly=True, size=(32, 1)),
        sg.InputText(key=HEADER, size=(32, 1)),
        sg.Multiline(key=TYPEDEF, enter_submits=False, autoscroll=True, enable_events=True, size=(64, 1))
    ]

    typedef_row = [ [ sg.Text(TYPEDEF, key=TYPEDEF+'-label'), sg.Push(), inputs[4]] ]
    struct_row = [ [ sg.Text(ELEMENTS), sg.Text('None', key='struct-fields', size=(48, 1)),
                     sg.Button('Edit elements', key='edit-elements', enable_events=True) ] ]
    enum_row = [ [ sg.Text(CONSTANTS), sg.Text('None', key='enum-values', size=(48, 1)),
                     sg.Button('Edit enum constants', key='edit-constants', enable_events=True) ] ]

    layout = [
        ([ sg.Text([ TYPE_NAME, DESCRIPTION, KIND, HEADER ][i], k=[ TYPE_NAME, DESCRIPTION, KIND, HEADER ][i]+'-label'),
           sg.Push(), inputs[i] ] for i in range(4)),
                [ hidable(typedef_row, key='typedef-row')],
                [ hidable(struct_row, key='struct-row') ],
                [ hidable(enum_row, key='enum-row')],
                [ sg.HorizontalSeparator() ],
                [ sg.Text('', key='status', size=(48,1), text_color='red'), sg.Push(),
                        sg.Button('Clear', key='clear'), sg.Button('Copy table to clipboard', key='copy'),
                        sg.Button('Close', key='close') ]
             ]

    clearable_keys = [ TYPE_NAME, DESCRIPTION, HEADER, TYPEDEF ] + [ 'struct-fields', 'enum-values']

    window = sg.Window('Fill in type table', layout, finalize=True)

    code_indent = '\n   '
    code_cell = inputs[4]
    code_cell.update(' .. code-block::' + code_indent)
    code_cell.bind("<Return>", "RETURN")

    window['struct-row'].update(visible=False)
    window['enum-row'].update(visible=False)

    struct_elements = [ ]
    enum_constants = [ ]

    simply_checked_keys = [ TYPE_NAME, DESCRIPTION, HEADER ]

    inner_table_col_width = [ 32, 32, 64]

    while True:
        event, values = window.read()
        #print(event, values)

        if event == sg.WIN_CLOSED:
            break

        if event == 'clear':
            for k in clearable_keys:
                window[k]('')
            code_cell.update(' .. code-block::' + code_indent)
            struct_elements = [ ]
            enum_constants = [ ]
            window['struct-fields']('None')
            window['enum-values']('None')
        elif 'copy' in event:

            if not _is_input_valid(values, window, simply_checked_keys):
                continue

            table_str = _render_table(values, struct_elements, enum_constants)
            table_str = tabhelper._decorate_table(table_str, indent_size, add_table_directive)

            print(table_str)
            clipboard.copy(table_str)
        elif 'RETURN' in event:
            code_cell.update(code_cell.get() + code_indent)
        elif 'Kind' in event:
            if values['Kind:'] == 'Structure':
                window['typedef-row'].update(visible=False)
                window['enum-row'].update(visible=False)
                window['struct-row'].update(visible=True)
            elif values['Kind:'] == 'Typedef':
                window['struct-row'].update(visible=False)
                window['enum-row'].update(visible=False)
                window['typedef-row'].update(visible=True)
            elif values['Kind:'] == 'Enumeration':
                window['struct-row'].update(visible=False)
                window['typedef-row'].update(visible=False)
                window['enum-row'].update(visible=True)
        elif event == 'edit-elements':
            struct_elements = show_inner_table(struct_elements, ['Type:', 'Name:', 'Description:'],
                                               required_cols=[0, 1, 2], col_width = inner_table_col_width)
            struct_fields = ''
            for row in range(len(struct_elements)):
                struct_fields += struct_elements[row][1] + ','
            if not struct_fields:
                struct_fields = 'None'
            else:
                struct_fields = struct_fields[:-1]  #strip last ','
                if len(struct_fields) > 64:
                    struct_fields = struct_fields[:64] + '...'
            window['struct-fields'].update(struct_fields)
        elif event == 'edit-constants':
            enum_constants = show_inner_table(enum_constants, ['Constant:', 'Initial value:', 'Description:'],
                                              required_cols=[0, 2], col_width=inner_table_col_width)
            enum_idents = ''
            for row in range(len(enum_constants)):
                enum_idents += enum_constants[row][0] + ','
            if not enum_idents:
                enum_idents = 'None'
            else:
                enum_idents = enum_idents[:-1]  #strip last ','
                if len(enum_idents) > 64:
                    enum_idents = enum_idents[:64] + '...'
            window['enum-values'].update(enum_idents)

        if 'close' in event:
            break

    window.close()

def _is_input_valid(values, window, simply_checked_keys) -> bool:

    is_input_valid = True

    # Simple checks.
    window['status'].update('')
    for k in simply_checked_keys:
        if not values[k]:
            window[k+'-label'].update(text_color='red')
            is_input_valid = False
        else:
            window[k+'-label'].update(text_color=sg.DEFAULT_TEXT_COLOR)

    if values[KIND] == 'Typedef':
        try:
            type_name = get_typedef_name(values[TYPEDEF])
            window[TYPEDEF+'-label'].update(text_color=sg.DEFAULT_TEXT_COLOR)
        except Exception as e:
            print('C parser exception: ', str(e))
            window['status'].update('Invalid syntax')
            window[TYPEDEF+'-label'].update(text_color='red')
            return False

        if type_name is None:
            window['status'].update('Invalid typedef syntax')
            window[TYPEDEF+'-label'].update(text_color='red')
            window[TYPE_NAME+'-label'].update(text_color='red')
            return False

        if type_name != values[TYPE_NAME]:
            window['status'].update('Type name mismatch')
            window[TYPEDEF+'-label'].update(text_color='red')
            window[TYPE_NAME+'-label'].update(text_color='red')
            return False
        else:
            window[TYPEDEF+'-label'].update(text_color=sg.DEFAULT_TEXT_COLOR)
            window[TYPE_NAME+'-label'].update(text_color=sg.DEFAULT_TEXT_COLOR)

    return is_input_valid


def _render_table(values, struct_elements, enum_constants):

    # Create the table.
    header = [TYPE_NAME, values[TYPE_NAME]]
    table = [ [ DESCRIPTION, values[DESCRIPTION] ],
              [ KIND, values[KIND] ],
              [ HEADER, values[HEADER] ] ]
    if values[KIND] == 'Typedef':
        table.extend([ [ TYPEDEF, values[TYPEDEF] ] ])
    elif values[KIND] == 'Structure':
        nested_table = None
        if len(struct_elements) > 0:
            nested_table = tabulate(struct_elements, tablefmt="grid")
            table.extend([ [ ELEMENTS, nested_table ] ])
        else:
            table.extend([ [ ELEMENTS, 'None' ] ])
    elif values[KIND] == 'Enumeration':
        nested_table = None
        if len(enum_constants) > 0:
            nested_table = tabulate(enum_constants, tablefmt="grid")
            table.extend([ [ CONSTANTS, nested_table ] ])
        else:
            table.extend([ [ CONSTANTS, 'None' ] ])

    table_str = tabulate(table, headers=header, tablefmt="grid")

    if values[KIND] == 'Structure' and nested_table != None:
        table_str = tabhelper.merge_nested_tables({ ELEMENTS : nested_table }, table_str)
    elif values[KIND] == 'Enumeration' and nested_table != None:
        table_str = tabhelper.merge_nested_tables({ CONSTANTS : nested_table }, table_str)

    return table_str




#
# destables
# Copyright (C) 2022  Arthur Wisz
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
#

import PySimpleGUI as sg
import re

def hidable(layout, key):
    return sg.pin(sg.Column(layout, key=key))

def show_inner_table(initial_values, header, required_cols, col_width):

    result = initial_values

    MAX_ROWS = 1 + 16
    MAX_COL = len(header)

    inner_layout = [ [ sg.Input(pad=(1,1), justification='left', key=str(i)+','+str(j), enable_events=True, size=(col_width[j], 1))
                      for j in range(MAX_COL) ] for i in range(MAX_ROWS) ]

    column_layout = [ [ hidable([ inner_layout[row] ], key=row) ] for row in range(MAX_ROWS) ]

    layout = [ [ sg.Col(column_layout, scrollable=True) ],
               [ sg.Push(), sg.Ok(size=(15, 1)), sg.Cancel(size=(15, 1)) ] ]

    window = sg.Window('Add feature elements', layout, finalize=True)

    #Fill header row.
    for col in range(MAX_COL):
        window['0,'+str(col)].update(background_color='cyan')   #doesn't work :(
        window['0,'+str(col)].update(value=header[col], disabled=True)

    for row in range(len(initial_values)):
        for col in range(MAX_COL):
            window[str(row+1)+','+str(col)].update(value=initial_values[row][col])

    for row in range(1, MAX_ROWS):
        if row > len(initial_values)+1:
            window[row].update(visible=False)
        for col in range(MAX_COL):
            window[str(row)+','+str(col)].bind("<Return>", "_RETURN")
    last_visible_row = len(initial_values) + 1
    window[str(last_visible_row)+',0'].set_focus()

    while True:
        event, values = window.read()

        if event != None and isinstance(event, str):
            m = re.match(r'([0-9]+),([0-9]+)_RETURN', event)
            if m != None:
                row = m.group(1)
                col = m.group(2)
                has_val = [ values[row+','+str(c)] != '' or c not in required_cols for c in range(MAX_COL) ]
                if all(has_val) and int(row) < MAX_ROWS-1:
                    #show next row if current row is filled in
                    last_visible_row = int(row)+1
                    window[last_visible_row].update(visible=True)
                    window[str(int(row)+1)+',0'].set_focus()

        if event == sg.WIN_CLOSED or event == 'Cancel':
            break

        if event == 'Ok':
            result = []
            for row in range(1, last_visible_row+1):
                col_values = []
                for col in range(MAX_COL):
                    col_values.append(values[str(row)+','+str(col)])
                #Required entries must not be empty.
                if all([ col_values[c] != '' or c not in required_cols for c in range(MAX_COL) ]):
                    result.append(col_values)
            break

    window.close()
    #print(result)
    return result



#
# destables
# Copyright (C) 2022  Arthur Wisz
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
#

import PySimpleGUI as sg
from commonwin import *
import clipboard
import tabhelper
from tabulate import tabulate
from c11tools import *

VAR_GROUP = 'Variables Group:'
HEADER = 'Declared in:'
VARIABLES = 'Variables:'
DESCRIPTION = 'Description:'
SYNTAX = 'Syntax:'

def show_variable_table(indent_size, add_table_directive):

    inputs = [
        sg.Input(key=VAR_GROUP, size=(48, 1), enable_events=True),
        sg.Input(key=HEADER, size=(32, 1), enable_events=True),
    ]

    layout = [
        ([ sg.Text([ VAR_GROUP, HEADER ][i], k=[ VAR_GROUP, HEADER ][i]+'-label'),
           sg.Push(), inputs[i] ] for i in range(2)),
        [ sg.Text(VARIABLES, k=VARIABLES+'-label'), sg.Text('', key='variables', size=(48, 1)), sg.Push(), sg.Button('Edit variables', key='edit-variables', enable_events=True) ],
        [ sg.HorizontalSeparator() ],
        [ sg.Text('', key='status', size=(48,1), text_color='red'), sg.Push(),
                sg.Button('Clear', key='clear'), sg.Button('Copy table to clipboard', key='copy'),
                sg.Button('Close', key='close') ]
    ]

    clearable_keys = [ VAR_GROUP, HEADER ]

    simply_checked_keys = [ VAR_GROUP, HEADER ]

    variables = []

    window = sg.Window('Fill in variables table', layout, finalize=True)

    while True:
        event, values = window.read()
        #print(event, values)

        if event == sg.WIN_CLOSED:
            break

        if event == 'clear':
            for k in clearable_keys:
                window[k]('')
            window['variables']('')
        elif 'copy' in event:
            window['status'].update('')

            if not _is_input_valid(values, window, simply_checked_keys, variables):
                continue

            table_str = _render_table(values, variables)
            table_str = tabhelper._decorate_table(table_str, indent_size, add_table_directive)

            print(table_str)
            clipboard.copy(table_str)
        elif event in [ 'edit-variables' ]:
            variables = show_inner_table(variables, [ SYNTAX, DESCRIPTION ],
                                                      required_cols=[0, 1], col_width = [ 48, 64 ])
            var_list = ''
            window['status'].update('')
            for row in range(len(variables)):
                try:
                    var_name = get_variable_name(variables[row][0])
                except Exception as e:
                    print('C parser exception: ', str(e))
                    window['status'].update('Invalid syntax')
                    var_name = ''
                var_list += var_name + ','
            if not var_list:
                var_list = 'None'
            else:
                var_list = var_list[:-1]  #strip last ','
                if len(var_list) > 64:
                    var_list = var_list[:64] + '...'
            window[event.replace('edit-', '')].update(var_list)

        if 'close' in event:
            break

    window.close()


def _is_input_valid(values, window, simply_checked_keys, vars) -> bool:

    is_input_valid = True
    window['status'].update('')
    for k in simply_checked_keys:
        if not values[k]:
            window[k+'-label'].update(text_color='red')
            is_input_valid = False
        else:
            window[k+'-label'].update(text_color=sg.DEFAULT_TEXT_COLOR)

    if not is_input_valid:
        return False

    for i in range(len(vars)):
        try:
            var_name = get_variable_name(vars[i][0])
        except Exception as e:
            print("error parsing '", vars[i][0], "': ", str(e))
            window['status'].update("Invalid syntax of variable " + str(i))
            return False

    return len(vars) > 0

def _render_table(values, vars):

    header = [ VAR_GROUP, values[VAR_GROUP] ]
    table = [ [ HEADER, values[HEADER]] ]

    var_tab = []
    for i in range(0, len(vars)):
        var_tab.append([ DESCRIPTION, vars[i][1] ])
        syntax = " .. code-block::\n\n    " + vars[i][0]
        var_tab.append([ SYNTAX, syntax ])

    nested_table = tabulate(var_tab, tablefmt="grid")
    table.extend([ [ VARIABLES, nested_table ] ])

    table_str = tabulate(table, headers=header, tablefmt="grid")
    table_str = tabhelper.merge_nested_tables({ VARIABLES : nested_table }, table_str)
    return table_str


#
# destables
# Copyright (C) 2022  Arthur Wisz
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
#

import tabulate
import re

def _find_header_rownum(header, tab_lines):
    header_row_num = None
    for l in tab_lines:
        if header in l:
            header_row_num = tab_lines.index(l)
            break
    if header_row_num is None:
        raise Exception('Inconsistent arguments: nested table header not found in table')
    return header_row_num

def _replace(s, at, line):
    return line[:at] + s + line[at+len(s):]

def merge_nested_tables(nested_dict, table_str):

    tab_lines = table_str.split('\n')
    for nested_header in nested_dict:
        nested_table = nested_dict[nested_header]
        nested_tab_lines = nested_table.split('\n')

        header_row_num = _find_header_rownum(nested_header, tab_lines)
        bottom_frame_num = header_row_num + 1
        while tab_lines[bottom_frame_num][0] == '|':
            bottom_frame_num += 1

        nested_tab_pos = tab_lines[header_row_num].find(nested_tab_lines[0])
        if nested_tab_pos < 0:
            raise Exception('Inconsistent arguments: 1st line of nested table not found in table')
        nested_tab_lines[0] = '-' + nested_tab_lines[0][1:-1] + '-'
        # top frame row
        tab_lines[header_row_num-1] = _replace(nested_tab_lines[0], nested_tab_pos, tab_lines[header_row_num-1])
        # bottom frame row
        tab_lines[bottom_frame_num] = _replace(nested_tab_lines[0], nested_tab_pos, tab_lines[bottom_frame_num])
        # copy nested table header to next row
        tab_lines[header_row_num+1] = _replace(nested_header, tab_lines[header_row_num].find(nested_header), tab_lines[header_row_num+1])
        # perform the merge
        for row in range(header_row_num, bottom_frame_num):
            tab_lines[row] = tab_lines[row].replace('| +', '+--')
            # For example '+     |' gets replaced with '------+'
            def repl_fun(matchobj):
                return '-' + '-' * len(matchobj.group(1)) + '+'
            tab_lines[row] = re.sub(r'\+([ ]+)\|', repl_fun, tab_lines[row])
            for i in range(len(tab_lines[header_row_num-1])):
                if tab_lines[row][i] == '|' and tab_lines[header_row_num-1][i] != '+':
                    tab_lines[row] = _replace(' ', i, tab_lines[row])
        # remove lower reduntant frame row
        del tab_lines[bottom_frame_num-1]
        # remove upper reduntant frame row
        del tab_lines[header_row_num]

        result = ''
        for l in tab_lines:
            result += l + '\n'
        return result

def _decorate_table(table_str, indent_size, add_table_dir):

    decorated_str = '.. table::\n\n' if add_table_dir else ''

    lines = table_str.split('\n')
    for l in lines:
        decorated_str += indent_size * ' ' + l + '\n'

    return decorated_str



#
# destables
# Copyright (C) 2022  Arthur Wisz
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
#

from antlr4 import *
from cparser.C11Lexer import C11Lexer
from cparser.C11Parser import C11Parser
from cparser.C11Visitor import C11Visitor

parser = None

def init_parser():

    global parser

    parser = C11Parser(None)

def _decl_from_syntax(syntax):
    decl = syntax.replace('.. code-block::', '')
    decl = decl.replace('\n', ' ')
    decl = decl.strip() + ';'
    return decl

def _parse_declaration(syntax):
    global parser

    c_decl = _decl_from_syntax(syntax)
    if c_decl[-1] != ';':
        c_decl.append(';')

    input_stream = InputStream(c_decl)

    lexer = C11Lexer(input_stream)
    stream = CommonTokenStream(lexer)
    parser.setInputStream(stream)
    parse_tree = parser.declaration()
    return parse_tree


class FuncVisitor(C11Visitor):

    def __init__(self) -> None:
        super().__init__()
        self.func_name = None
        self.func_params = []
        self._in_params_list = False

    def visitDirectDeclarator(self, ctx: C11Parser.DirectDeclaratorContext):

        if not self._in_params_list:
            if ctx.LeftParen() != None and ctx.RightParen() != None and ctx.parameterTypeList() != None:
                self.func_name = ctx.directDeclarator().Identifier().symbol.text
        else:
            if ctx.Identifier() != None:
                self.func_params.append(ctx.Identifier().symbol.text)
        return super().visitDirectDeclarator(ctx)

    def visitParameterList(self, ctx: C11Parser.ParameterListContext):
        self._in_params_list = True
        return super().visitParameterList(ctx)


def get_func_info(syntax):

    parse_tree = _parse_declaration(syntax)

    visitor = FuncVisitor()
    visitor.visit(tree=parse_tree)
    return visitor.func_name, visitor.func_params


class TypedefVisitor(C11Visitor):

    def __init__(self) -> None:
        super().__init__()
        self.typedef_name = ''
        self.is_typedef = False
        self._skip_identifiers = False

    def visitStorageClassSpecifier(self, ctx: C11Parser.StorageClassSpecifierContext):
        if ctx.Typedef() != None:
            self.is_typedef = True
        return super().visitStorageClassSpecifier(ctx)

    def visitDirectDeclarator(self, ctx: C11Parser.DirectDeclaratorContext):
        if not self._skip_identifiers:
            if ctx.Identifier() is not None:
                self.typedef_name = ctx.Identifier().symbol.text
        return super().visitDirectDeclarator(ctx)

    def visitTypeSpecifier(self, ctx: C11Parser.TypeSpecifierContext):
        if not self._skip_identifiers:
            if ctx.typedefName() != None:
                self.typedef_name = ctx.typedefName().Identifier().symbol.text
        return super().visitTypeSpecifier(ctx)

    #Happens for function type typedefs.
    def visitParameterList(self, ctx: C11Parser.ParameterListContext):
        self._skip_identifiers = True
        return super().defaultResult()


def get_typedef_name(syntax):

    parse_tree = _parse_declaration(syntax)

    visitor = TypedefVisitor()
    visitor.visit(tree=parse_tree)

    if visitor.is_typedef:
        return visitor.typedef_name
    return None

class VarVisitor(C11Visitor):

    def __init__(self) -> None:
        super().__init__()

    def visitDirectDeclarator(self, ctx: C11Parser.DirectDeclaratorContext):
        if ctx.Identifier() is not None:
            self.var_name = ctx.Identifier().symbol.text
        return super().visitDirectDeclarator(ctx)

def get_variable_name(syntax):

    parse_tree = _parse_declaration(syntax)

    visitor = VarVisitor()
    visitor.visit(tree=parse_tree)
    return visitor.var_name



# Generated from C11.g4 by ANTLR 4.9.3
from antlr4 import *
if __name__ is not None and "." in __name__:
    from .C11Parser import C11Parser
else:
    from C11Parser import C11Parser

# This class defines a complete generic visitor for a parse tree produced by C11Parser.

class C11Visitor(ParseTreeVisitor):

    # Visit a parse tree produced by C11Parser#primaryExpression.
    def visitPrimaryExpression(self, ctx:C11Parser.PrimaryExpressionContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by C11Parser#genericSelection.
    def visitGenericSelection(self, ctx:C11Parser.GenericSelectionContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by C11Parser#genericAssocList.
    def visitGenericAssocList(self, ctx:C11Parser.GenericAssocListContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by C11Parser#genericAssociation.
    def visitGenericAssociation(self, ctx:C11Parser.GenericAssociationContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by C11Parser#postfixExpression.
    def visitPostfixExpression(self, ctx:C11Parser.PostfixExpressionContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by C11Parser#argumentExpressionList.
    def visitArgumentExpressionList(self, ctx:C11Parser.ArgumentExpressionListContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by C11Parser#unaryExpression.
    def visitUnaryExpression(self, ctx:C11Parser.UnaryExpressionContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by C11Parser#unaryOperator.
    def visitUnaryOperator(self, ctx:C11Parser.UnaryOperatorContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by C11Parser#castExpression.
    def visitCastExpression(self, ctx:C11Parser.CastExpressionContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by C11Parser#multiplicativeExpression.
    def visitMultiplicativeExpression(self, ctx:C11Parser.MultiplicativeExpressionContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by C11Parser#additiveExpression.
    def visitAdditiveExpression(self, ctx:C11Parser.AdditiveExpressionContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by C11Parser#shiftExpression.
    def visitShiftExpression(self, ctx:C11Parser.ShiftExpressionContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by C11Parser#relationalExpression.
    def visitRelationalExpression(self, ctx:C11Parser.RelationalExpressionContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by C11Parser#equalityExpression.
    def visitEqualityExpression(self, ctx:C11Parser.EqualityExpressionContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by C11Parser#andExpression.
    def visitAndExpression(self, ctx:C11Parser.AndExpressionContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by C11Parser#exclusiveOrExpression.
    def visitExclusiveOrExpression(self, ctx:C11Parser.ExclusiveOrExpressionContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by C11Parser#inclusiveOrExpression.
    def visitInclusiveOrExpression(self, ctx:C11Parser.InclusiveOrExpressionContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by C11Parser#logicalAndExpression.
    def visitLogicalAndExpression(self, ctx:C11Parser.LogicalAndExpressionContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by C11Parser#logicalOrExpression.
    def visitLogicalOrExpression(self, ctx:C11Parser.LogicalOrExpressionContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by C11Parser#conditionalExpression.
    def visitConditionalExpression(self, ctx:C11Parser.ConditionalExpressionContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by C11Parser#assignmentExpression.
    def visitAssignmentExpression(self, ctx:C11Parser.AssignmentExpressionContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by C11Parser#assignmentOperator.
    def visitAssignmentOperator(self, ctx:C11Parser.AssignmentOperatorContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by C11Parser#expression.
    def visitExpression(self, ctx:C11Parser.ExpressionContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by C11Parser#constantExpression.
    def visitConstantExpression(self, ctx:C11Parser.ConstantExpressionContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by C11Parser#declaration.
    def visitDeclaration(self, ctx:C11Parser.DeclarationContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by C11Parser#declarationSpecifiers.
    def visitDeclarationSpecifiers(self, ctx:C11Parser.DeclarationSpecifiersContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by C11Parser#declarationSpecifiers2.
    def visitDeclarationSpecifiers2(self, ctx:C11Parser.DeclarationSpecifiers2Context):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by C11Parser#declarationSpecifier.
    def visitDeclarationSpecifier(self, ctx:C11Parser.DeclarationSpecifierContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by C11Parser#initDeclaratorList.
    def visitInitDeclaratorList(self, ctx:C11Parser.InitDeclaratorListContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by C11Parser#initDeclarator.
    def visitInitDeclarator(self, ctx:C11Parser.InitDeclaratorContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by C11Parser#storageClassSpecifier.
    def visitStorageClassSpecifier(self, ctx:C11Parser.StorageClassSpecifierContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by C11Parser#typeSpecifier.
    def visitTypeSpecifier(self, ctx:C11Parser.TypeSpecifierContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by C11Parser#structOrUnionSpecifier.
    def visitStructOrUnionSpecifier(self, ctx:C11Parser.StructOrUnionSpecifierContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by C11Parser#structOrUnion.
    def visitStructOrUnion(self, ctx:C11Parser.StructOrUnionContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by C11Parser#structDeclarationList.
    def visitStructDeclarationList(self, ctx:C11Parser.StructDeclarationListContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by C11Parser#structDeclaration.
    def visitStructDeclaration(self, ctx:C11Parser.StructDeclarationContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by C11Parser#specifierQualifierList.
    def visitSpecifierQualifierList(self, ctx:C11Parser.SpecifierQualifierListContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by C11Parser#structDeclaratorList.
    def visitStructDeclaratorList(self, ctx:C11Parser.StructDeclaratorListContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by C11Parser#structDeclarator.
    def visitStructDeclarator(self, ctx:C11Parser.StructDeclaratorContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by C11Parser#enumSpecifier.
    def visitEnumSpecifier(self, ctx:C11Parser.EnumSpecifierContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by C11Parser#enumeratorList.
    def visitEnumeratorList(self, ctx:C11Parser.EnumeratorListContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by C11Parser#enumerator.
    def visitEnumerator(self, ctx:C11Parser.EnumeratorContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by C11Parser#enumerationConstant.
    def visitEnumerationConstant(self, ctx:C11Parser.EnumerationConstantContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by C11Parser#atomicTypeSpecifier.
    def visitAtomicTypeSpecifier(self, ctx:C11Parser.AtomicTypeSpecifierContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by C11Parser#typeQualifier.
    def visitTypeQualifier(self, ctx:C11Parser.TypeQualifierContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by C11Parser#functionSpecifier.
    def visitFunctionSpecifier(self, ctx:C11Parser.FunctionSpecifierContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by C11Parser#alignmentSpecifier.
    def visitAlignmentSpecifier(self, ctx:C11Parser.AlignmentSpecifierContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by C11Parser#declarator.
    def visitDeclarator(self, ctx:C11Parser.DeclaratorContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by C11Parser#directDeclarator.
    def visitDirectDeclarator(self, ctx:C11Parser.DirectDeclaratorContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by C11Parser#gccDeclaratorExtension.
    def visitGccDeclaratorExtension(self, ctx:C11Parser.GccDeclaratorExtensionContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by C11Parser#gccAttributeSpecifier.
    def visitGccAttributeSpecifier(self, ctx:C11Parser.GccAttributeSpecifierContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by C11Parser#gccAttributeList.
    def visitGccAttributeList(self, ctx:C11Parser.GccAttributeListContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by C11Parser#gccAttribute.
    def visitGccAttribute(self, ctx:C11Parser.GccAttributeContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by C11Parser#nestedParenthesesBlock.
    def visitNestedParenthesesBlock(self, ctx:C11Parser.NestedParenthesesBlockContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by C11Parser#pointer.
    def visitPointer(self, ctx:C11Parser.PointerContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by C11Parser#typeQualifierList.
    def visitTypeQualifierList(self, ctx:C11Parser.TypeQualifierListContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by C11Parser#parameterTypeList.
    def visitParameterTypeList(self, ctx:C11Parser.ParameterTypeListContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by C11Parser#parameterList.
    def visitParameterList(self, ctx:C11Parser.ParameterListContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by C11Parser#parameterDeclaration.
    def visitParameterDeclaration(self, ctx:C11Parser.ParameterDeclarationContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by C11Parser#identifierList.
    def visitIdentifierList(self, ctx:C11Parser.IdentifierListContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by C11Parser#typeName.
    def visitTypeName(self, ctx:C11Parser.TypeNameContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by C11Parser#abstractDeclarator.
    def visitAbstractDeclarator(self, ctx:C11Parser.AbstractDeclaratorContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by C11Parser#directAbstractDeclarator.
    def visitDirectAbstractDeclarator(self, ctx:C11Parser.DirectAbstractDeclaratorContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by C11Parser#typedefName.
    def visitTypedefName(self, ctx:C11Parser.TypedefNameContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by C11Parser#initializer.
    def visitInitializer(self, ctx:C11Parser.InitializerContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by C11Parser#initializerList.
    def visitInitializerList(self, ctx:C11Parser.InitializerListContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by C11Parser#designation.
    def visitDesignation(self, ctx:C11Parser.DesignationContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by C11Parser#designatorList.
    def visitDesignatorList(self, ctx:C11Parser.DesignatorListContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by C11Parser#designator.
    def visitDesignator(self, ctx:C11Parser.DesignatorContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by C11Parser#staticAssertDeclaration.
    def visitStaticAssertDeclaration(self, ctx:C11Parser.StaticAssertDeclarationContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by C11Parser#statement.
    def visitStatement(self, ctx:C11Parser.StatementContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by C11Parser#labeledStatement.
    def visitLabeledStatement(self, ctx:C11Parser.LabeledStatementContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by C11Parser#compoundStatement.
    def visitCompoundStatement(self, ctx:C11Parser.CompoundStatementContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by C11Parser#blockItemList.
    def visitBlockItemList(self, ctx:C11Parser.BlockItemListContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by C11Parser#blockItem.
    def visitBlockItem(self, ctx:C11Parser.BlockItemContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by C11Parser#expressionStatement.
    def visitExpressionStatement(self, ctx:C11Parser.ExpressionStatementContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by C11Parser#selectionStatement.
    def visitSelectionStatement(self, ctx:C11Parser.SelectionStatementContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by C11Parser#iterationStatement.
    def visitIterationStatement(self, ctx:C11Parser.IterationStatementContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by C11Parser#forCondition.
    def visitForCondition(self, ctx:C11Parser.ForConditionContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by C11Parser#forDeclaration.
    def visitForDeclaration(self, ctx:C11Parser.ForDeclarationContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by C11Parser#forExpression.
    def visitForExpression(self, ctx:C11Parser.ForExpressionContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by C11Parser#jumpStatement.
    def visitJumpStatement(self, ctx:C11Parser.JumpStatementContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by C11Parser#compilationUnit.
    def visitCompilationUnit(self, ctx:C11Parser.CompilationUnitContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by C11Parser#translationUnit.
    def visitTranslationUnit(self, ctx:C11Parser.TranslationUnitContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by C11Parser#externalDeclaration.
    def visitExternalDeclaration(self, ctx:C11Parser.ExternalDeclarationContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by C11Parser#functionDefinition.
    def visitFunctionDefinition(self, ctx:C11Parser.FunctionDefinitionContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by C11Parser#declarationList.
    def visitDeclarationList(self, ctx:C11Parser.DeclarationListContext):
        return self.visitChildren(ctx)



del C11Parser


#
# destables
# Copyright (C) 2022  Arthur Wisz
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
#

import PySimpleGUI as sg
from commonwin import *
import clipboard
import tabhelper
from tabulate import tabulate
from c11tools import *

FUNC_NAME = 'Function name:'
DESCRIPTION = 'Description:'
SYNTAX = 'Syntax:'
HEADER = 'Declared in:'
ISR = 'May be called from ISR:'
REENTRANT = 'Reentrancy:'
RETURN = 'Return value:'
IN_PARAMS = 'Parameters [in]:'
OUT_PARAMS = 'Parameters [out]:'
IN_OUT_PARAMS = 'Parameters [in-out]:'
CALL_CYCLE = 'Call cycle interval:'


def show_function_table(indent_size, add_table_directive):

    inputs = [
        sg.Input(key=FUNC_NAME, size=(48, 1), enable_events=True),
        sg.Multiline(key=DESCRIPTION, enter_submits=False, autoscroll=True, size=(64, 2), enable_events=True),
        sg.Multiline(key=SYNTAX, enter_submits=False, autoscroll=True, enable_events=True, size=(64, 5)),
        sg.Input(key=HEADER, size=(32, 1), enable_events=True),
    ]

    layout = [
        ([ sg.Text([ FUNC_NAME, DESCRIPTION, SYNTAX, HEADER ][i], k=[ FUNC_NAME, DESCRIPTION, SYNTAX, HEADER ][i]+'-label'),
           sg.Push(), inputs[i] ] for i in range(4)),
        [ sg.Checkbox(text=ISR, key=ISR) ],
        [ sg.Checkbox(text=REENTRANT, key=REENTRANT) ],
        [ sg.Text(RETURN, k=RETURN+'-label'), sg.Push(), sg.Input('void', key='retval-type', size=(32,1)), sg.Input(key='retval-description', size=(64,1)) ],
        [ sg.Text(IN_PARAMS, k=IN_PARAMS+'-label'), sg.Text('None', key='in-params', size=(48, 1)), sg.Push(), sg.Button('Edit input parameters', key='edit-in-params', enable_events=True) ],
        [ sg.Text(OUT_PARAMS, k=OUT_PARAMS+'-label'), sg.Text('None', key='out-params', size=(48, 1)), sg.Push(), sg.Button('Edit output parameters', key='edit-out-params', enable_events=True) ],
        [ sg.Text(IN_OUT_PARAMS, k=IN_OUT_PARAMS+'-label'), sg.Text('None', key='inout-params', size=(48, 1)), sg.Push(), sg.Button('Edit in-out parameters', key='edit-inout-params', enable_events=True) ],
        [ sg.Checkbox(text='Is cyclic', key='is-cyclic', enable_events=True), sg.Push(), sg.Text(CALL_CYCLE, k=CALL_CYCLE+'-label'), sg.Input(key=CALL_CYCLE, enable_events=True) ],
        [ sg.HorizontalSeparator() ],
        [ sg.Text('', key='status', size=(48,1), text_color='red'), sg.Push(),
                sg.Button('Clear', key='clear'), sg.Button('Copy table to clipboard', key='copy'),
                sg.Button('Close', key='close') ]
    ]

    clearable_keys = [ FUNC_NAME, DESCRIPTION, SYNTAX, HEADER, ISR, REENTRANT, RETURN,
                       IN_PARAMS, OUT_PARAMS, IN_OUT_PARAMS, CALL_CYCLE ]

    simply_checked_keys = [ FUNC_NAME, DESCRIPTION, HEADER ]

    window = sg.Window('Fill in function table', layout, finalize=True)

    code_indent = '\n   '
    code_cell = inputs[2]
    code_cell.update(' .. code-block::' + code_indent)
    code_cell.bind("<Return>", "RETURN")

    window[CALL_CYCLE].update(disabled=True)

    params_dict = {
        'in-params': [],
        'out-params' : [],
        'inout-params' : []
    }

    while True:
        event, values = window.read()
        #print(event, values)

        if event == sg.WIN_CLOSED:
            break

        if event == 'clear':
            for k in clearable_keys:
                window[k]('')
            code_cell.update(' .. code-block::' + code_indent)
            for k in params_dict:
                params_dict[k] = []
            window['in-params']('None')
            window['out-params']('None')
            window['inout-params']('None')
        elif 'copy' in event:

            if not _is_input_valid(values, window, simply_checked_keys, params_dict):
                continue

            table_str = _render_table(values, params_dict)
            table_str = tabhelper._decorate_table(table_str, indent_size, add_table_directive)

            print(table_str)
            clipboard.copy(table_str)
        elif 'RETURN' in event:
            code_cell.update(code_cell.get() + code_indent)
        elif event in [ 'edit-in-params', 'edit-out-params', 'edit-inout-params' ]:
            params_key = event.replace('edit-', '')
            params_dict[params_key] = show_inner_table(params_dict[params_key], ['Parameter:', 'Description:'],
                                                       required_cols=[0, 1], col_width = [ 32, 64 ])
            params_list = ''
            for row in range(len(params_dict[params_key])):
                params_list += params_dict[params_key][row][0] + ','
            if not params_list:
                params_list = 'None'
            else:
                params_list = params_list[:-1]  #strip last ','
                if len(params_list) > 64:
                    params_list = params_list[:64] + '...'
            window[event.replace('edit-', '')].update(params_list)
        elif event == 'is-cyclic':
            if not values['is-cyclic']:
                window[CALL_CYCLE].update('')
            window[CALL_CYCLE].update(disabled=not values['is-cyclic'])
        elif event in simply_checked_keys + [ SYNTAX, CALL_CYCLE ]:
            window[event + '-label'].update(text_color=sg.DEFAULT_TEXT_COLOR)

        if 'close' in event:
            break

    window.close()


def _is_input_valid(values, window, simply_checked_keys, params_dict) -> bool:

    is_input_valid = True

    # Simple checks.
    window['status'].update('')
    for k in simply_checked_keys:
        if not values[k]:
            window[k+'-label'].update(text_color='red')
            is_input_valid = False
        else:
            window[k+'-label'].update(text_color=sg.DEFAULT_TEXT_COLOR)

    if values['is-cyclic'] and not values[CALL_CYCLE]:
        window[CALL_CYCLE+'-label'].update(text_color='red')
        is_input_valid = False
    else:
        window[CALL_CYCLE+'-label'].update(text_color=sg.DEFAULT_TEXT_COLOR)

    if values['retval-type'] != 'void' and not values['retval-description']:
        window[RETURN+'-label'].update(text_color='red')
        is_input_valid = False
    else:
        window[RETURN+'-label'].update(text_color=sg.DEFAULT_TEXT_COLOR)

    if not is_input_valid:
        return False

    # Check consistency of the input.
    try:
        func_name, func_params = get_func_info(values[SYNTAX])
        window[SYNTAX+'-label'].update(text_color=sg.DEFAULT_TEXT_COLOR)
    except Exception as e:
        print('C parser exception: ', str(e))
        window['status'].update('Invalid syntax')
        window[SYNTAX+'-label'].update(text_color='red')
        return False

    if values[FUNC_NAME] == '' or func_name != values[FUNC_NAME].strip():
        window['status'].update('Function name mismatch')
        window[SYNTAX+'-label'].update(text_color='red')
        window[FUNC_NAME+'-label'].update(text_color='red')
        return False
    else:
        window[SYNTAX+'-label'].update(text_color=sg.DEFAULT_TEXT_COLOR)
        window[FUNC_NAME+'-label'].update(text_color=sg.DEFAULT_TEXT_COLOR)

    documented_params = []
    for k in params_dict:
        if len(params_dict[k]) > 0:
            for par in params_dict[k]:
                documented_params.append(par[0])
    undocumented_params = list(set(func_params) - set(documented_params))

    if len(undocumented_params) > 0:
        window['status'].update(str(len(undocumented_params)) + ' undocumented parameter(s)')
        return False

    params_missing_in_syntax = list(set(documented_params) - set(func_params))
    if len(params_missing_in_syntax) > 0:
        window[SYNTAX+'-label'].update(text_color='red')
        window['status'].update(str(len(params_missing_in_syntax)) + ' missing parameter(s) in function syntax')
        return False

    return True

def _render_table(values, params_dict):

    # Create the table.
    header = [ FUNC_NAME, values[FUNC_NAME]]
    table = [ [ DESCRIPTION, values[DESCRIPTION ] ],
                [ SYNTAX, values[SYNTAX] ],
                [ HEADER, values[HEADER] ]  ]
    table.append( [ ISR, 'Yes' if values[ISR] else 'No' ] )
    table.append( [ REENTRANT, 'Reentrant' if values[REENTRANT] else 'Non-reentrant' ] )

    nested_return_table = None
    if values['retval-type'] != 'void':
        retval = [ [ values['retval-type'], values['retval-description'] ] ]
        nested_return_table = tabulate(retval, tablefmt="grid")
        table.append( [ RETURN, nested_return_table ] )
    else:
        table.append( [ RETURN, 'None' ] )

    nested_params_tables = { 'in-params': None, 'out-params' : None, 'inout-params' : None   }
    params_headers = { 'in-params': IN_PARAMS, 'out-params' : OUT_PARAMS, 'inout-params' : IN_OUT_PARAMS }

    for i in [ 'in-params', 'out-params', 'inout-params' ]:
        if len(params_dict[i]) > 0:
            nested_params_tables[i] = tabulate(params_dict[i], tablefmt="grid")
            table.append([ params_headers[i], nested_params_tables[i] ])
        #else:
        #    table.append([ params_headers[i], 'None' ])

    if values['is-cyclic']:
        table.append( [ CALL_CYCLE, values[CALL_CYCLE] ] )

    table_str = tabulate(table, headers=header, tablefmt="grid")

    if nested_return_table != None:
        table_str = tabhelper.merge_nested_tables({ RETURN : nested_return_table }, table_str)

    for i in [ 'in-params', 'out-params', 'inout-params' ]:
        if nested_params_tables[i] != None:
            table_str = tabhelper.merge_nested_tables({ params_headers[i] : nested_params_tables[i] }, table_str)

    return table_str



#
# destables
# Copyright (C) 2022  Arthur Wisz
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
#

import os
import sys
sys.path.append(os.path.dirname(os.path.abspath(__file__)))

from typewin import *
from funcwin import *
from varwin import *
import PySimpleGUI as sg
from c11tools import init_parser

TABLE_DIRECTIVE = 'Add .. table:: directive'
INDENT_SIZE = "Indentation"

def main():

    init_parser()

    layout = [ [ sg.Checkbox(text=TABLE_DIRECTIVE, key=TABLE_DIRECTIVE, default=True) ],
               [ sg.Text(INDENT_SIZE, k=INDENT_SIZE+'-label'), sg.Push(),
                 sg.Slider(orientation ='horizontal', key=INDENT_SIZE, range=(0,4), default_value=4, size=(15, 20)) ],
              [ sg.HorizontalSeparator() ],
              [ sg.Button('Type table', key='type', size = (30, 2)) ],
              [ sg.Button('Function table', key='function', size = (30, 2)) ],
              [ sg.Button('Variables table', key='variables', size = (30, 2)) ],
#              [ sg.Button('Macro constants table', key='macro-constants', size = (30, 2)) ],
#              [ sg.Button('Macro function table', key='macro-function', size = (30, 2)) ],
              [ sg.HorizontalSeparator() ],
              [ sg.Button('Exit', size = (30, 3))] ]

    window = sg.Window('SW Design Tables Tool', layout)

    while True:
        event, values = window.read()
        print(event, values)
        indent_size = int(values[INDENT_SIZE])
        add_table_dir = values[TABLE_DIRECTIVE]
        if event == sg.WIN_CLOSED or event == 'Exit':
            break
        if event == 'type':
            show_type_table(indent_size, add_table_dir)
        elif event == 'function':
            show_function_table(indent_size, add_table_dir)
        elif event == 'variables':
            show_variable_table(indent_size, add_table_dir)

    window.close()

if __name__ == "__main__":
    main()



from setuptools import setup

# Metadata goes in setup.cfg. These are here for GitHub's dependency graph.
setup(
    name="destables",
    install_requires=[
        'PySimpleGUI>=4.56.0',
        'tabulate>=0.8.9',
        'clipboard>=0.0.4',
        'antlr4-python3-runtime==4.9.3'
    ],
    extras_require={
    },
)
